@@ -2,36 +2,22 @@ package Log::Any::Adapter::TAP;
 use strict;
 use warnings;
 use parent 'Log::Any::Adapter::Base';
+use Log::Any ();
 use Try::Tiny;
 use Carp 'croak';
 require Scalar::Util;
 require Data::Dumper;
 
-our $VERSION= '0.002000';
+our $VERSION= '0.003000';
 
 # ABSTRACT: Logging adapter suitable for use in TAP testcases
 
 
-our $global_filter_level;    # default for level-filtering
-our %category_filter_level;  # per-category filter levels
-our $show_category;          # whether to show logging category on each message
-our $show_file_line;         # Whether to show caller for each message
-our $show_file_fullname;     # whether to use full path for caller info
-our $show_usage;             # whether to print usage notes on initialization
 our %level_map;              # mapping from level name to numeric level
-
-sub _coerce_filter_level {
-	my $val= shift;
-	return (!defined $val || $val eq 'none')? $level_map{trace}-1
-		: ($val eq 'all')? $level_map{emergency}
-		: exists $level_map{$val}? $level_map{$val}
-		: ($val =~ /^([A-Za-z]+)([-+][0-9]+)$/) && defined $level_map{lc $1}? $level_map{lc $1} + $2
-		: croak "unknown log level '$val'";
-}
-
 BEGIN {
 	# Initialize globals, and use %ENV vars for defaults
 	%level_map= (
+        min       => -1,
 		trace     => -1,
 		debug     =>  0,
 		info      =>  1,
@@ -41,8 +27,8 @@ BEGIN {
 		critical  =>  5,
 		alert     =>  6,
 		emergency =>  7,
+        max       =>  7,
 	);
-	
 	# Make sure we have numeric levels for all the core logging methods
 	for ( Log::Any->logging_methods() ) {
 		if (!defined $level_map{$_}) {
@@ -57,7 +43,26 @@ BEGIN {
 	my %aliases= Log::Any->log_level_aliases;
 	$level_map{$_} ||= $level_map{$aliases{$_}}
 		for keys %aliases;
-	
+}
+
+sub _log_level_value { $level_map{$_[1]} }
+
+sub _coerce_filter_level {
+	my $val= shift;
+	return (!defined $val || $val eq 'none')? $level_map{trace}-1
+		: ($val eq 'all')? $level_map{emergency}
+		: exists $level_map{$val}? $level_map{$val}
+		: ($val =~ /^([A-Za-z]+)([-+][0-9]+)$/) && defined $level_map{lc $1}? $level_map{lc $1} + $2
+		: croak "unknown log level '$val'";
+}
+
+our $global_filter_level;    # default for level-filtering
+our %category_filter_level;  # per-category filter levels
+our $show_category;          # whether to show logging category on each message
+our $show_file_line;         # Whether to show caller for each message
+our $show_file_fullname;     # whether to use full path for caller info
+our $show_usage;             # whether to print usage notes on initialization
+BEGIN {
 	# Suppress debug and trace by default
 	$global_filter_level= 'debug';
 	
@@ -99,8 +104,10 @@ sub dumper { $_[0]{dumper} ||= $_[0]->default_dumper }
 sub category { $_[0]{category} }
 
 
+our $_show_dumper_warning= 1;
 sub init {
 	my $self= shift;
+	my $custom_dumper= $self->{dumper};
 	# Apply default dumper if not set
 	$self->{dumper} ||= $self->default_dumper;
 	# Apply default filter if not set
@@ -126,30 +133,54 @@ sub init {
 		           ." all log levels, and TAP_LOG_ORIGIN=3 to see caller info.");
 		$show_usage= 0;
 	}
+	if ($custom_dumper && $_show_dumper_warning) {
+		$self->notice("Custom 'dumper' will not work with Log::Any versions >= 0.9");
+		$_show_dumper_warning= 0;
+	}
 	
 	return $self;
 }
 
 
+my %_tap_method;
 sub write_msg {
 	my ($self, $level_name, $str)= @_;
+	
 	chomp $str;
-	$str =~ s/\n/\n#   /sg;
+	$str= "$level_name: $str" unless $level_name eq 'info';
+	
 	if ($show_category) {
 		$str .= ' (' . $self->category . ')';
 	}
+	
 	if ($show_file_line) {
 		my $i= 0;
-		++$i while caller($i) =~ '^Log::Any';
+		++$i while caller($i) =~ /^Log::Any(:|$)/;
 		my (undef, $file, $line)= caller($i);
 		$file =~ s|.*/lib/||
 			unless $show_file_fullname;
 		$str .= ' (' . $file . ':' . $line . ')';
 	}
-	if ($level_map{$level_name} >= $level_map{warning}) {
-		print STDERR ($level_name eq 'info'? '# ' : "# $level_name: "), $str, "\n";
-	} else {
-		print STDOUT ($level_name eq 'info'? '# ' : "# $level_name: "), $str, "\n";
+
+	# Was going to cache more of this, but logger might load before Test::More,
+	# so better to keep testing it each time.  At least cache which method name we're using.
+	my $name= ($_tap_method{$level_name} ||=
+		($self->_log_level_value($level_name) >= $self->_log_level_value('warning')?
+			'diag':'note'));
+	my $m;
+	if ($m= main->can($name)) {
+		$m->($str);
+	}
+	elsif (Test::Builder->can('new')) {
+		Test::Builder->new->$name($str);
+	}
+	else {
+		$str =~ s/\n/\n#   /sg;
+		if ($name eq 'diag') {
+			print STDERR "# $str\n";
+		} else {
+			print STDOUT "# $str\n";
+		}
 	}
 }
 
@@ -247,9 +278,21 @@ sub _build_filtered_subclasses {
 	}
 }
 
+our $_called_as_fatal;
 BEGIN {
 	__PACKAGE__->_build_logging_methods;
 	__PACKAGE__->_build_filtered_subclasses;
+	
+	if ($Log::Any::VERSION >= 0.9 && $Log::Any::VERSION <= 1.03) {
+		# Log::Any broke the adapter contract a bit during these releases.
+		# This is an ugly hack to preserve the function of this module.
+		require Log::Any::Proxy;
+		no warnings 'redefine';
+		my $fatal= Log::Any::Proxy->can('fatal');
+		*Log::Any::Proxy::fatal= sub { local $_called_as_fatal= 1; $fatal->(@_) };
+		my $crit= \&critical;
+		*critical= sub { $_called_as_fatal? fatal(@_) : $crit->(@_) };
+	}
 }
 
 1;
@@ -266,7 +309,7 @@ Log::Any::Adapter::TAP - Logging adapter suitable for use in TAP testcases
 
 =head1 VERSION
 
-version 0.002000
+version 0.003000
 
 =head1 DESCRIPTION
 
@@ -347,18 +390,14 @@ so C<debug-1> is equivalent to C<trace> and C<debug+1> is equivalent to C<info>.
 This differs from syslog, where increasing numbers are less important.
 (why did they choose that??)
 
-=head2 dumper
+=head2 dumper (DEPRECATED, unusable in Log::Any >= 0.9)
 
   use Log::Any::Adapter 'TAP', dumper => sub { my $val=shift; ... };
 
-Use a custom dumper function for converting perl data to strings.
-The dumper is only used for the C<${level}f(...)> formatting functions,
-and for log levels C<debug> and C<trace>.
-All other logging will stringify the object in the normal way.
-
-Defaults to L</default_dumper>, which prints the data in "some human-readable
-format".  The default will NOT give you a pure serialization, and is subject
-to change.
+This feature lets you use a custom dumper in the printf-style logging
+functions.  However, these are no longer handled by the adapter in
+new versions of Log::Any, so you need to use a custom Proxy class in
+your log-producing module.
 
 =head1 METHODS
 
@@ -390,6 +429,15 @@ change to a more friendly format in the future.
 
 This module has all the standard logging methods from L<Log::Any/LOG LEVELS>.
 
+Note that the regular logging methods are only specified to take a single string.
+This module in the past supported passing objects as additional parameters, and
+having them stringified with a custom dumper, caatching exceptions thrown during
+stringification.  With the new Log::Any design, these things are decided in the
+producing module, so these features are no longer possible.
+
+If this module does receive multiple arguments or have its printf-formatting
+methods called, it does the following:
+
 For regular logging functions (i.e. C<warn>, C<info>) the arguments are
 stringified and concatenated.  Errors during stringify or printing are not
 caught.
@@ -409,7 +457,7 @@ Michael Conrad <mike@nrdvana.net>
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is copyright (c) 2014 by Michael Conrad.
+This software is copyright (c) 2015 by Michael Conrad.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
